"""
Module implementing an enhanced string column type for SQLAlchemy
with a support for regular expression operators in Postgres and SQLite.
"""

# courtesy of Xion: http://xion.io/post/code/sqlalchemy-regex-filters.html
import re
import sqlite3

from sqlalchemy import String as _String
from sqlalchemy import event, exc
from sqlalchemy.engine import Engine
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql.expression import BinaryExpression, func, literal
from sqlalchemy.sql.operators import custom_op

__all__ = ["String"]


class String(_String):
    """Enchanced version of standard SQLAlchemy's :class:`String`.

    Supports additional operators that can be used while constructing
    filter expressions.
    """

    class comparator_factory(_String.comparator_factory):
        """Contains implementation of :class:`String` operators
        related to regular expressions.
        """

        def regexp(self, other):
            return RegexMatchExpression(self.expr, literal(other), custom_op("~"))

        def iregexp(self, other):
            return RegexMatchExpression(self.expr, literal(other), custom_op("~*"))

        def not_regexp(self, other):
            return RegexMatchExpression(self.expr, literal(other), custom_op("!~"))

        def not_iregexp(self, other):
            return RegexMatchExpression(self.expr, literal(other), custom_op("!~*"))


class RegexMatchExpression(BinaryExpression):
    """Represents matching of a column againsts a regular expression."""


@compiles(RegexMatchExpression, "sqlite")
def sqlite_regex_match(element, compiler, **kw):
    """Compile the SQL expression representing a regular expression match
    for the SQLite engine.
    """
    # determine the name of a custom SQLite function to use for the operator
    operator = element.operator.opstring
    try:
        func_name, _ = SQLITE_REGEX_FUNCTIONS[operator]
    except (KeyError, ValueError) as e:
        would_be_sql_string = " ".join(
            (compiler.process(element.left), operator, compiler.process(element.right))
        )
        raise exc.StatementError(
            "unknown regular expression match operator: %s" % operator,
            would_be_sql_string,
            None,
            e,
        )

    # compile the expression as an invocation of the custom function
    regex_func = getattr(func, func_name)
    regex_func_call = regex_func(element.left, element.right)
    return compiler.process(regex_func_call)


@event.listens_for(Engine, "connect")
def sqlite_engine_connect(dbapi_connection, connection_record):
    """Listener for the event of establishing connection to a SQLite database.

    Creates the functions handling regular expression operators
    within SQLite engine, pointing them to their Python implementations above.
    """
    if not isinstance(dbapi_connection, sqlite3.Connection):
        return

    for name, function in SQLITE_REGEX_FUNCTIONS.values():
        dbapi_connection.create_function(name, 2, function)


# Mapping from the regular expression matching operators
# to named Python functions that implement them for SQLite.
SQLITE_REGEX_FUNCTIONS = {
    "~": ("REGEXP", lambda value, regex: bool(re.match(regex, value))),
    "~*": ("IREGEXP", lambda value, regex: bool(re.match(regex, value, re.IGNORECASE))),
    "!~": ("NOT_REGEXP", lambda value, regex: not re.match(regex, value)),
    "!~*": (
        "NOT_IREGEXP",
        lambda value, regex: not re.match(regex, value, re.IGNORECASE),
    ),
}



import asyncio
import logging
import os
import queue
import threading

from flask_script import Command, Manager, Option

from aardvark import create_app
from aardvark.exceptions import AardvarkException
from aardvark.configuration import CONFIG, create_config
from aardvark.persistence.sqlalchemy import SQLAlchemyPersistence
from aardvark.retrievers.runner import RetrieverRunner

manager = Manager(create_app)
log = logging.getLogger("aardvark")

ACCOUNT_QUEUE = queue.Queue()
QUEUE_LOCK = threading.Lock()
UPDATE_DONE = False

SWAG_REPO_URL = "https://github.com/Netflix-Skunkworks/swag-client"

LOCALDB = "sqlite"
DEFAULT_LOCALDB_FILENAME = "aardvark.db"

# Configuration default values.
DEFAULT_SWAG_BUCKET = "swag-data"
DEFAULT_AARDVARK_ROLE = "Aardvark"
DEFAULT_NUM_THREADS = 5


# All of these default to None rather than the corresponding DEFAULT_* values
# so we can tell whether they were passed or not. We don't prompt for any of
# the options that were passed as parameters.
@manager.option("-a", "--aardvark-role", dest="aardvark_role_param", type=str)
@manager.option("-b", "--swag-bucket", dest="bucket_param", type=str)
@manager.option("-d", "--db-uri", dest="db_uri_param", type=str)
@manager.option("--num-threads", dest="num_threads_param", type=int)
@manager.option("--no-prompt", dest="no_prompt", action="store_true", default=False)
def config(
    aardvark_role_param, bucket_param, db_uri_param, num_threads_param, no_prompt
):
    """
    Creates a config.py configuration file from user input or default values.

    If all configurable values are specified by parameters, user input
    is not needed and will not be prompted.

    If the no-prompt flag is not set, user input will be prompted for
    each of the configurable values not specified by parameters.

    If the no-prompt flag is set, no user input will be collected and
    the configuration file will be populated with option-specified values
    or defaults.

    The resulting configuration file defines the following parameters.
    Configurable parameters are shown in <angle braces>.

    SWAG_OPTS = {'swag.type': 's3', 'swag.bucket_name': <bucket>}
    SWAG_FILTER = None
    SWAG_SERVICE_ENABLED_REQUIREMENT = None
    ROLENAME = <aardvark_role>
    REGION = "us-east-1"
    SQLALCHEMY_DATABASE_URI = <db_uri>
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    NUM_THREADS = <num_threads>
    LOG_CFG = {...}
    """
    # We don't set these until runtime.
    default_db_uri = f"{LOCALDB}:///{os.getcwd()}/{DEFAULT_LOCALDB_FILENAME}"

    if no_prompt:  # Just take the parameters as currently constituted.
        aardvark_role = aardvark_role_param or DEFAULT_AARDVARK_ROLE
        num_threads = num_threads_param or DEFAULT_NUM_THREADS
        db_uri = db_uri_param or default_db_uri

        # If a swag bucket was specified we set write_swag here so it gets
        # written out to the config file below.
        bucket = bucket_param or DEFAULT_SWAG_BUCKET

    else:
        # This is essentially the same "param, or input, or default"
        # structure as the additional parameters below.
        if bucket_param:
            bucket = bucket_param
        else:
            print(f"\nAardvark can use SWAG to look up accounts. See {SWAG_REPO_URL}")
            use_swag = input("Do you use SWAG to track accounts? [yN]: ")
            if len(use_swag) > 0 and "yes".startswith(use_swag.lower()):
                bucket_prompt = f"SWAG_BUCKET [{DEFAULT_SWAG_BUCKET}]: "
                bucket = input(bucket_prompt) or DEFAULT_SWAG_BUCKET
            else:
                bucket = ""

        aardvark_role_prompt = f"ROLENAME [{DEFAULT_AARDVARK_ROLE}]: "
        db_uri_prompt = f"DATABASE URI [{default_db_uri}]: "
        num_threads_prompt = f"# THREADS [{DEFAULT_NUM_THREADS}]: "

        aardvark_role = (
            aardvark_role_param or input(aardvark_role_prompt) or DEFAULT_AARDVARK_ROLE
        )
        db_uri = db_uri_param or input(db_uri_prompt) or default_db_uri
        num_threads = (
            num_threads_param or input(num_threads_prompt) or DEFAULT_NUM_THREADS
        )

    create_config(
        aardvark_role=aardvark_role,
        swag_bucket=bucket or "",
        swag_filter="",
        swag_service_enabled_requirement="",
        sqlalchemy_database_uri=db_uri,
        sqlalchemy_track_modifications=False,
        num_threads=num_threads,
        region="us-east-1",
    )


@manager.option("-a", "--accounts", dest="accounts", type=str, default="all")
@manager.option("-r", "--arns", dest="arns", type=str, default="all")
def update(accounts, arns):
    """
    Asks AWS for new Access Advisor information.
    """
    # The runner will default to all accounts and ARNs if None is passed in
    accounts = None if accounts == "all" else accounts.split(",")
    arns = None if arns == "all" else arns.split(",")

    r = RetrieverRunner()
    try:
        asyncio.run(r.run(accounts=accounts, arns=arns))
    except KeyboardInterrupt:
        r.cancel()
    except AardvarkException as e:
        log.error(e)
        exit(1)


class GunicornServer(Command):
    """
    This is the main GunicornServer server, it runs the flask app with gunicorn and
    uses any configuration options passed to it.
    You can pass all standard gunicorn flags to this command as if you were
    running gunicorn itself.
    For example:
    aardvark start_api -w 4 -b 127.0.0.0:8002
    Will start gunicorn with 4 workers bound to 127.0.0.0:8002
    """

    description = "Run the app within Gunicorn"

    def get_options(self):
        options = []
        try:
            from gunicorn.config import make_settings
        except ImportError:
            # Gunicorn does not yet support Windows.
            # See issue #524. https://github.com/benoitc/gunicorn/issues/524
            # For dev on Windows, make this an optional import.
            print("Could not import gunicorn, skipping.")
            return options

        settings = make_settings()
        for setting, klass in settings.items():
            if klass.cli:
                if klass.action:
                    if klass.action == "store_const":
                        options.append(
                            Option(*klass.cli, const=klass.const, action=klass.action)
                        )
                    else:
                        options.append(Option(*klass.cli, action=klass.action))
                else:
                    options.append(Option(*klass.cli))
        return options

    def run(self, *args, **kwargs):
        from gunicorn.app.wsgiapp import WSGIApplication

        app = WSGIApplication()

        app.app_uri = "aardvark:create_app()"
        return app.run()


@manager.command
def drop_db():
    """ Drops the database. """
    SQLAlchemyPersistence().teardown_db()


@manager.command
def create_db():
    """ Creates the database. """
    SQLAlchemyPersistence().init_db()


def main():
    manager.add_command("start_api", GunicornServer())
    manager.run()


if __name__ == "__main__":
    main()



from sqlalchemy import TIMESTAMP, BigInteger, Column, ForeignKey, Integer, Text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

from aardvark.utils.sqla_regex import String

Base = declarative_base()


class AdvisorData(Base):
    """
    Models certain IAM Access Advisor Data fields.

    {
      "totalAuthenticatedEntities": 1,
      "lastAuthenticatedEntity": "arn:aws:iam::XXXXXXXX:role/name",
      "serviceName": "Amazon Simple Systems Manager",
      "lastAuthenticated": 1489176000000,
      "serviceNamespace": "ssm"
    }
    """

    __tablename__ = "advisor_data"
    id = Column(Integer, primary_key=True)
    item_id = Column(
        Integer, ForeignKey("aws_iam_object.id"), nullable=False, index=True
    )
    lastAuthenticated = Column(BigInteger)
    serviceName = Column(String(128), index=True)
    serviceNamespace = Column(String(64), index=True)
    lastAuthenticatedEntity = Column(Text)
    totalAuthenticatedEntities = Column(Integer)


class AWSIAMObject(Base):
    """
    Meant to model AWS IAM Object Access Advisor.
    """

    __tablename__ = "aws_iam_object"
    id = Column(Integer, primary_key=True)
    arn = Column(String(2048), nullable=True, index=True, unique=True)
    lastUpdated = Column(TIMESTAMP)
    usage = relationship(
        "AdvisorData",
        backref="item",
        cascade="all, delete, delete-orphan",
        foreign_keys="AdvisorData.item_id",
    )
    action_usage = relationship(
        "ActionData",
        backref="item",
        cascade="all, delete, delete-orphan",
        foreign_keys="ActionData.item_id",
    )


class ActionData(Base):
    """
    Models action-specific data from sources other than Access Advisor,
    such as CloudTrail.
    """

    __tablename__ = "action_data"
    id = Column(Integer, primary_key=True)
    item_id = Column(
        Integer, ForeignKey("aws_iam_object.id"), nullable=False, index=True
    )
    lastAuthenticated = Column(BigInteger)
    serviceName = Column(String(128), index=True)
    serviceNamespace = Column(String(64), index=True)



# ensure absolute import for python3
from __future__ import absolute_import

import logging
import os.path
from logging.config import dictConfig

from flasgger import Swagger
from flask import Flask

from aardvark.configuration import CONFIG
from aardvark.persistence.sqlalchemy import SQLAlchemyPersistence
from aardvark.view import advisor_bp

BLUEPRINTS = [advisor_bp]

API_VERSION = "1"

persistence = SQLAlchemyPersistence()
dictConfig(CONFIG["logging"].get())
log = logging.getLogger("aardvark")


def create_app(test_config=None):
    app = Flask(__name__, static_url_path="/static")
    Swagger(app)

    if test_config is not None:
        app.config.update(test_config)

    # For ELB and/or Eureka
    @app.route("/healthcheck")
    def healthcheck():
        """Healthcheck
        Simple healthcheck that indicates the services is up
        ---
        responses:
          200:
            description: service is up
        """
        return "ok"

    # Blueprints
    for bp in BLUEPRINTS:
        app.register_blueprint(bp, url_prefix="/api/{0}".format(API_VERSION))

    # Extensions:
    persistence.init_db()

    return app


def _find_config():
    """Search for config.py in order of preference and return path if it exists, else None"""
    CONFIG_PATHS = [
        os.path.join(os.getcwd(), "config.py"),
        "/etc/aardvark/config.py",
        "/apps/aardvark/config.py",
    ]
    for path in CONFIG_PATHS:
        if os.path.exists(path):
            return path
    return None



class AardvarkException(Exception):
    pass


class AccessAdvisorException(AardvarkException):
    pass


class CombineException(AardvarkException):
    pass


class DatabaseException(AardvarkException):
    pass


class RetrieverException(AardvarkException):
    pass



import asyncio
import logging
import re
from copy import copy
from typing import Any, Dict, List

import confuse
from asgiref.sync import sync_to_async
from botocore.exceptions import ClientError
from cloudaux.aws.iam import list_roles, list_users
from cloudaux.aws.sts import boto3_cached_conn
from swag_client import InvalidSWAGDataException
from swag_client.backend import SWAGManager
from swag_client.util import parse_swag_config_options

from aardvark.exceptions import RetrieverException
from aardvark.plugins import AardvarkPlugin
from aardvark.persistence.sqlalchemy import SQLAlchemyPersistence
from aardvark.retrievers import RetrieverPlugin
from aardvark.retrievers.access_advisor import AccessAdvisorRetriever

log = logging.getLogger("aardvark")
sap = SQLAlchemyPersistence()
re_account_id = re.compile(r"\d{12}")


class RetrieverRunner(AardvarkPlugin):
    """Scheduling and execution for data retrieval tasks."""

    retrievers: List[RetrieverPlugin]
    account_queue: asyncio.Queue
    arn_queue: asyncio.Queue
    results_queue: asyncio.Queue
    failed_arns: List[str]
    tasks: List[asyncio.Future]
    num_workers: int
    swag: SWAGManager
    swag_config: confuse.ConfigView

    def __init__(
        self,
        alternative_config: confuse.Configuration = None,
    ):
        super().__init__(alternative_config=alternative_config)
        self.tasks = []
        self.retrievers = []
        self.failed_arns = []
        self.num_workers = self.config["updater"]["num_threads"].as_number()
        self.swag_config = self.config["swag"]
        swag_opts = parse_swag_config_options(self.swag_config["opts"].get())
        self.swag = SWAGManager(**swag_opts)

    def register_retriever(self, r: RetrieverPlugin):
        """Add a retriever instance to be called during the run process."""
        self.retrievers.append(r)

    async def _run_retrievers(self, arn: str) -> Dict[str, Any]:
        """Run retriever plugins for a given ARN.

        Retriever plugins are executed in the order in which they are registered. Each retriever
        is passed the result from the previous one, starting with a dict containing an `arn` element.

        Note: The data from all previous retriever plugins is mutable by subsequent ones.
        """
        data = {
            "arn": arn,
        }
        # Iterate through retrievers, passing the results from the previous to the next.
        for r in self.retrievers:
            try:
                data = await r.run(arn, data)
            except Exception as e:
                log.error(f"failed to run {r} on ARN {arn}")
                raise RetrieverException from e
        return data

    async def _retriever_loop(self, name: str):
        """Loop to consume from self.arn_queue and call the retriever runner function."""
        log.debug(f"creating {name}")
        while True:
            log.debug("getting arn from queue")
            arn = await self.arn_queue.get()
            log.debug(f"{name} retrieving data for {arn}")
            try:
                data = await self._run_retrievers(arn)
            except Exception as e:
                log.error(f"failed to run retriever on ARN {arn}: {e}")
                self.failed_arns.append(arn)
                self.arn_queue.task_done()
                continue
            # TODO: handle nested data from retrievers in persistence layer
            await self.results_queue.put(data)
            self.arn_queue.task_done()

    async def _results_loop(self, name: str):
        """Loop to consume from self.results_queue and handle results."""
        log.debug(f"creating {name}")
        while True:
            data = await self.results_queue.get()
            log.debug(f"{name} storing results for {data['arn']}")
            await sync_to_async(sap.store_role_data)(
                {data["arn"]: data["access_advisor"]}
            )
            self.results_queue.task_done()

    async def _get_arns_for_account(self, account: str):
        """Retrieve ARNs for roles, users, policies, and groups in an account and add them to the ARN queue."""
        conn_details: Dict[str, str] = {
            "account_number": account,
            "assume_role": self.config["aws"]["rolename"].as_str(),
            "session_name": "aardvark",
            "region": self.config["aws"]["region"].as_str() or "us-east-1",
            "arn_partition": self.config["aws"]["arn_partition"].as_str() or "aws",
        }
        client = await sync_to_async(boto3_cached_conn)(
            "iam", service_type="client", **conn_details
        )

        for role in await sync_to_async(list_roles)(**conn_details):
            await self.arn_queue.put(role["Arn"])

        for user in await sync_to_async(list_users)(**conn_details):
            await self.arn_queue.put(user["Arn"])

        for page in await sync_to_async(client.get_paginator("list_policies").paginate)(
            Scope="Local"
        ):
            for policy in page["Policies"]:
                await self.arn_queue.put(policy["Arn"])

        for page in await sync_to_async(client.get_paginator("list_groups").paginate)():
            for group in page["Groups"]:
                await self.arn_queue.put(group["Arn"])

    async def _arn_lookup_loop(self, name: str):
        """Loop to consume from self.account_queue to retrieve and enqueue ARNs for each account."""
        log.debug(f"creating {name}")
        while True:
            account = await self.account_queue.get()
            log.debug(f"{name} retrieving ARNs for {account}")
            await self._get_arns_for_account(account)
            self.account_queue.task_done()

    async def _get_swag_accounts(self) -> List[Dict]:
        """Retrieve AWS accounts from SWAG based on the SWAG options in the application configuration."""
        log.debug("getting accounts from SWAG")
        try:
            all_accounts: List[Dict] = self.swag.get_all(
                self.swag_config["filter"].get()
            )
            swag_service = self.swag_config["service_enabled_requirement"].get()
            if swag_service:
                all_accounts = await sync_to_async(self.swag.get_service_enabled)(
                    swag_service, accounts_list=all_accounts
                )
        except (KeyError, InvalidSWAGDataException, ClientError) as e:
            log.error(
                f"Account names passed but SWAG not configured or unavailable: {e}"
            )
            raise RetrieverException("Could not retrieve SWAG data") from e

        return all_accounts

    async def _queue_all_accounts(self):
        """Add all accounts to the account queue.

        Perform a SWAG lookup and add all returned accounts to `self.account_queue`."""
        for account in await self._get_swag_accounts():
            await self.account_queue.put(account["id"])

    async def _queue_arns(self, arns: List[str]):
        """Add a list of ARNs to the ARN queue."""
        for arn in arns:
            await self.arn_queue.put(arn)

    async def _queue_accounts(self, account_names: List[str]):
        """Add requested accounts to the account queue.

        Given a list of account names and/or IDs, use SWAG to look up account numbers where needed
        and add each account number to `self.account_queue`."""
        accounts = copy(account_names)
        for account in accounts:
            if re_account_id.match(account):
                await self.account_queue.put(account)
                accounts.remove(account)

        all_accounts = await self._get_swag_accounts()

        # TODO(psanders): Consider refactoring. This could be expensive for organizations
        #  with many accounts and many aliases.
        for account in all_accounts:
            # Check if the account name matches one we want. If so, queue it and carry on.
            if account.get("name") in accounts:
                await self.account_queue.put(account["id"])
                continue
            # Now check the account's aliases to see if one matches.
            alias_key = "aliases" if account["schemaVersion"] == "2" else "alias"
            for alias in account.get(alias_key, []):
                if alias in accounts:
                    await self.account_queue.put(account["id"])
                    continue

    def cancel(self):
        """Send a cancel signal to all running workers."""
        log.info("Stopping runner tasks")
        for task in self.tasks:
            task.cancel()
            log.info(f"Task {task} canceled")

    async def run(self, accounts: List[str] = None, arns: List[str] = None):
        """Prep account queue and kick off ARN lookup, retriever, and results workers.

        Populate ARN queue with ARNs if provided. Otherwise use SWAG to look up account
        numbers and put those in the account queue.

        After that, we start `updater.num_threads` workers for each queue. Workers will NOT
        be started for the account queue if ARNs are provided since there will be no accounts
        in the queue.
        """
        self.register_retriever(AccessAdvisorRetriever())
        log.debug("starting retriever")

        self.arn_queue = asyncio.Queue()
        self.account_queue = asyncio.Queue()
        self.results_queue = asyncio.Queue()

        lookup_accounts = True
        if arns:
            await self._queue_arns(arns)
            lookup_accounts = False

        # We only need to do account lookups if ARNs were not provided.
        if lookup_accounts:
            if accounts:
                await self._queue_accounts(accounts)
            else:
                await self._queue_all_accounts()

            for i in range(self.num_workers):
                name = f"arn-lookup-worker-{i}"
                task = asyncio.create_task(self._arn_lookup_loop(name))
                self.tasks.append(task)

        for i in range(self.num_workers):
            name = f"retriever-worker-{i}"
            task = asyncio.create_task(self._retriever_loop(name))
            self.tasks.append(task)

        for i in range(self.num_workers):
            name = f"results-worker-{i}"
            task = asyncio.create_task(self._results_loop(name))
            self.tasks.append(task)

        await self.account_queue.join()
        await self.arn_queue.join()
        await self.results_queue.join()

        # Clean up our workers
        self.cancel()

        await asyncio.gather(*self.tasks, return_exceptions=True)
